__version__ = "0.2.3"
__license__ = "GPLv3"

__author__ = "José Antonio Perdiguero López"
__email__ = "perdy.hh@gmail.com"

__url__ = "https://github.com/PeRDy/apistar-crud"
__description__ = "API Star tools to create CRUD resources."



import re
import typing

from apistar import Route


class BaseResource(type):
    METHODS = {
        "list": ("/", "GET"),  # List resource collection
        "drop": ("/", "DELETE"),  # Drop resource entire collection
        "create": ("/", "POST"),  # Create a new element for this resource
        "retrieve": ("/{element_id}/", "GET"),  # Retrieve an element of this resource
        "update": ("/{element_id}/", "PUT"),  # Update an element of this resource
        "delete": ("/{element_id}/", "DELETE"),  # Delete an element of this resource
    }
    AVAILABLE_METHODS = tuple(METHODS.keys())
    DEFAULT_METHODS = ("create", "retrieve", "update", "delete", "list")

    def __new__(mcs, name, bases, namespace):
        try:
            resource_name = namespace.get("name", name.lower())
            model = namespace["model"]
            input_type = namespace["input_type"]
            output_type = namespace["output_type"]
        except KeyError as e:
            raise AttributeError('{} needs to define attribute: "{}"'.format(name, e))

        # Check resource name validity
        if re.match("[a-zA-Z][-_a-zA-Z]", resource_name) is None:
            raise AttributeError('Invalid resource name "{}"'.format(resource_name))

        # Add valid name and verbose name
        namespace["name"] = resource_name
        namespace["verbose_name"] = namespace.get("verbose_name", namespace["name"])

        # Default columns and order
        namespace["columns"] = namespace.get("columns", ["id"])
        namespace["order"] = namespace.get("order", "id")

        methods = namespace.get("methods", mcs.DEFAULT_METHODS)

        # Create CRUD methods and routes
        mcs.add_methods(namespace, methods, model, input_type, output_type)
        mcs.add_routes(namespace, methods)

        return type(name, bases, namespace)

    @classmethod
    def add_routes(mcs, namespace: typing.Dict[str, typing.Any], methods: typing.Iterable[str]):
        class Routes:
            """Routes descriptor"""

            def __get__(self, instance, owner):
                return [Route(*mcs.METHODS[method], getattr(owner, method), name=method) for method in methods]

        namespace["routes"] = Routes()

    @classmethod
    def add_methods(
        mcs, namespace: typing.Dict[str, typing.Any], methods: typing.Iterable[str], model, input_type, output_type
    ):
        # Generate CRUD methods
        crud_ns = {
            k: v for m in methods for k, v in getattr(mcs, "add_{}".format(m))(model, input_type, output_type).items()
        }

        # Preserve already defined methods
        crud_ns.update({m: crud_ns[f"_{m}"] for m in methods if m not in namespace})

        namespace.update(crud_ns)

    @classmethod
    def add_create(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        raise NotImplementedError

    @classmethod
    def add_retrieve(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        raise NotImplementedError

    @classmethod
    def add_update(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        raise NotImplementedError

    @classmethod
    def add_delete(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        raise NotImplementedError

    @classmethod
    def add_list(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        raise NotImplementedError

    @classmethod
    def add_drop(mcs, model, input_type, output_type) -> typing.Dict[str, typing.Any]:
        raise NotImplementedError



import typing
from collections import namedtuple

from apistar import Include

from apistar_crud.admin import Admin

__all__ = ["routes"]

RouteOptions = namedtuple("RouteOptions", ("url", "admin"))


class Routes:
    def __init__(self):
        self.resources = {}

    def register(self, resource, url: str = None, admin: bool = True):
        """
        Register a resource.

        :param resource: Resource.
        :param url: Route url.
        :param admin: True if should be added to admin site.
        """
        url = url or "/{}".format(resource.name)
        self.resources[resource] = RouteOptions(url, admin)

    def routes(self, admin="/admin") -> typing.List[Include]:
        """
        Generate the list of routes for all resources and admin views.

        :param admin: Admin path, disabled if None.
        :return: List of routes.
        """
        r = [Include(opts.url, r.name, r.routes) for r, opts in self.resources.items()]

        private_routes = []

        if admin:
            a = Admin(*[r for r, opts in self.resources.items() if opts.admin])
            r.append(Include(admin, "admin", a.routes, documented=False))
            private_routes += a.metadata_routes

        if private_routes:
            r.append(Include("/_crud", "crud", private_routes, documented=False))

        return r


routes = Routes()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['apistar_crud', 'apistar_crud.resource']

package_data = \
{'': ['*'], 'apistar_crud': ['static/*', 'templates/*']}

install_requires = \
['apistar-pagination>=0.3.0,<0.4.0', 'apistar>=0.5.41,<0.6.0']

setup_kwargs = {
    'name': 'apistar-crud',
    'version': '0.6.4',
    'description': 'API Star tools to create CRUD resources.',
    'long_description': '# API Star CRUD\n[![Build Status](https://travis-ci.org/PeRDy/apistar-crud.svg?branch=master)](https://travis-ci.org/PeRDy/apistar-crud)\n[![codecov](https://codecov.io/gh/PeRDy/apistar-crud/branch/master/graph/badge.svg)](https://codecov.io/gh/PeRDy/apistar-crud)\n[![PyPI version](https://badge.fury.io/py/apistar-crud.svg)](https://badge.fury.io/py/apistar-crud)\n\n* **Version:** 0.6.4\n* **Status:** Production/Stable\n* **Author:** José Antonio Perdiguero López\n\n## Table of Contents\n\n- [API Star CRUD](#api-star-crud)\n  * [Features](#features)\n    + [Resource](#resource)\n    + [ORM](#orm)\n    + [Admin site](#admin-site)\n  * [Quick start](#quick-start)\n  * [Usage](#usage)\n    + [SQLAlchemy](#sqlalchemy)\n    + [Peewee](#peewee)\n  * [Resources](#resources)\n    + [Routes](#routes)\n    + [Override methods](#override-methods)\n    + [Filtering](#filtering)\n  * [Admin](#admin)\n    + [Main page](#main-page)\n    + [Resource list](#resource-list)\n    + [Resource detail](#resource-detail)\n\n## Features\nAPI Star CRUD provides an easy way to define a REST resource and generic operations over it.\n\n### Resource\nThe resources are classes with a default implementation for **methods**:\n* `create`: Create a new element for this resource.\n* `retrieve`: Retrieve an element of this resource.\n* `update`: Update (partially or fully) an element of this resource.\n* `delete`: Delete an element of this resource.\n* `list`: List resource collection with **filtering** capabilities.\n* `drop`: Drop resource collection.\n\n### ORM\nAPI Star CRUD supports the following ORM:\n* [SQLAlchemy](https://www.sqlalchemy.org/) through [apistar-sqlalchemy](https://github.com/PeRDy/apistar-sqlalchemy).\n* [Peewee](https://github.com/coleifer/peewee) through [apistar-peewee-orm](https://github.com/PeRDy/apistar-peewee-orm).\n\n### Admin site\nAPI Star CRUD serves an admin site to handle resources in a graphical way, by default this site is routed to `/admin`. \n\n## Quick start\nInstall API Star CRUD:\n\n```bash\n$ pip install apistar-crud[peewee]\n```\n\nor \n\n```\n$ pip install apistar-crud[sqlalchemy]\n```\n\nFollow the steps:\n\n1. Create an **input type** and **output type** for your resource:\n2. Define a **model** based on your ORM.\n3. Build your **resource** using the metaclass specific for your ORM.\n4. Add the **routes** for your resources.\n\n## Usage\n### SQLAlchemy\nExample of a fully functional resource based on SQLAlchemy.\n\nCreate an **input type** and **output type**:\n\n```python\nclass PuppyInputType(types.Type):\n    name = validators.String()\n\nclass PuppyOutputType(types.Type):\n    id = validators.Integer()\n    name = validators.String()\n```\n\nDefine a **model**:\n\n```python\nclass PuppyModel(Base):\n    __tablename__ = "Puppy"\n\n    id = Column(Integer, primary_key=True)\n    name = Column(String)\n```\n\nThe **resource**:\n\n```python\nfrom apistar_crud.resource.sqlalchemy import Resource\n\nclass PuppyResource(metaclass=Resource):\n    name = "puppy"\n    verbose_name = "Puppy"\n    \n    model = PuppyModel\n    input_type = PuppyInputType\n    output_type = PuppyOutputType\n    methods = ("create", "retrieve", "update", "delete", "list", "drop")\n```\n\nThe resource generates his own **routes**:\n\n```python\nfrom apistar import App\nfrom apistar_crud.routes import routes as resource_routes\n\nresource_routes.register(PuppyResource, "/puppy", admin=True)\n\nroutes = [\n    # ... your app routes\n]\n\nroutes += resource_routes.routes(admin="/admin")\npackages = ("apistar_crud",)\napp = App(routes=routes, packages=packages)\n```\n\n### Peewee\nExample of a fully functional resource based on Peewee.\n\nCreate an **input type** and **output type**:\n\n```python\nclass PuppyInputType(types.Type):\n    name = validators.String()\n\nclass PuppyOutputType(types.Type):\n    id = validators.Integer()\n    name = validators.String()\n```\n\nDefine a **model**:\n\n```python\nclass PuppyModel(peewee.Model):\n    name = peewee.CharField()\n```\n\nThe **resource**:\n\n```python\nfrom apistar_crud.resource.peewee import Resource\n\nclass PuppyResource(metaclass=Resource):\n    name = "puppy"\n    verbose_name = "Puppy"\n    \n    model = PuppyModel\n    input_type = PuppyInputType\n    output_type = PuppyOutputType\n    methods = ("create", "retrieve", "update", "delete", "list", "drop")\n```\n\nThe resource generates his own **routes**:\n\n```python\nfrom apistar import App\nfrom apistar_crud.routes import routes as resource_routes\n\nresource_routes.register(PuppyResource, "/puppy", admin=True)\n\nroutes = [\n    # ... your app routes\n]\n\nroutes += resource_routes.routes(admin="/admin")\npackages = ("apistar_crud",)\napp = App(routes=routes, packages=packages)\n```\n\n## Resources\n\n### Routes\nThe routes for resource methods are:\n\n|Method  |Verb  |URL\n|--------|------|--------------\n|create  |POST  |/\n|retrieve|GET   |/{element_id}/\n|update  |PUT   |/{element_id}/\n|delete  |DELETE|/{element_id}/\n|list    |GET   |/\n|drop    |DELETE|/\n\n\n### Override methods\nTo customize CRUD methods you can override them like:\n\n```python\nfrom apistar_crud.resource.peewee import Resource\n\nclass PuppyResource(metaclass=Resource):\n    name = "puppy"\n    verbose_name = "Puppy"\n    \n    model = PuppyModel\n    input_type = PuppyInputType\n    output_type = PuppyOutputType\n    methods = ("create", "retrieve", "update", "delete", "list", "drop")\n    \n    @classmethod\n    def create(cls, element: PuppyInputType) -> PuppyOutputType:\n        # Do your custom process\n```\n\n### Filtering\nDefault `list` implementation has filtering capabilities but to reflect it in the schema is necessary to override that \nmethod defining all the parameters that will be used as a filter and pass it as keywords to `_list()` method. \n\nIt is a direct translation to ORM keyword arguments \nso in case of SQLAlchemy it uses a `filter_by()` method and a `where()` in Peewee case.\n\n```python\nimport typing\n\nfrom apistar import http\nfrom apistar_crud.resource.peewee import Resource\nfrom apistar_pagination import PageNumberResponse\n\nclass PuppyResource(metaclass=Resource):\n    name = "puppy"\n    verbose_name = "Puppy"\n    \n    model = PuppyModel\n    input_type = PuppyInputType\n    output_type = PuppyOutputType\n    methods = ("create", "retrieve", "update", "delete", "list", "drop")\n    \n    @classmethod\n    def list(cls, name: http.QueryParam, page: http.QueryParam=None, page_size: http.QueryParam=None) -> typing.List[PuppyOutputType]:\n        return PageNumberResponse(page=page, page_size=page_size, content=cls._filter(name=name))\n```\n\n## Admin\nAPI Star CRUD provides an admin panel to manage resources, including filtering, editing and creating new entries in your \nresources.\n\nTo include admin panel in your application you simply has to add the routes as following:\n\n```python\nfrom apistar import App\nfrom apistar_crud.routes import routes as resource_routes\n\nresource_routes.register(PuppyResource, "/puppy", admin=True)\n\nroutes = [\n    # ... your app routes\n]\n\nroutes += resource_routes.routes(admin="/admin")\npackages = ("apistar_crud",)\napp = App(routes=routes, packages=packages)\n```\n\n### Main page\nMain page showing a list of manageable resources.\n\n![Admin main page](docs/images/admin_landing.png "Main page showing a list of manageable resources")\n\n### Resource list\nConfigurable view of a list of resource entries. If that resource provides filtering parameters for list method, a \nfilter input will be available. The list of columns shown is also configurable.\n\n![Admin resource list page](docs/images/admin_resource_list.png "Configurable view of a list of resource entries")\n\n### Resource detail\nPage for editing a resource entry or creating a new one.\n\n![Admin resource detail page](docs/images/admin_resource_detail.png "Page for editing a resource entry or create a new one")\n',
    'author': 'José Antonio Perdiguero López',
    'author_email': 'perdy@perdy.io',
    'url': 'https://github.com/PeRDy/apistar-crud',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
